'''
Wrapper over FTDI API
'''

from pyftdi.ftdi import Ftdi
from pyftdi.spi import SpiController
from pyftdi.gpio import GpioAsyncController


class FtdiDevice:
    """FTDI FT2232H SPI master to access FPGA Control/Status Registers (CSR)
    plus some GPIO control.

    SPI parameters:
      - FTDI channel B:
        * BDBUS0 - SCLK
        * BDBUS1 - MOSI
        * BDBUS2 - MISO
        * BDBUS3 - CSn
      - slave;
      - mode 0 only;
      - most significant bit transmitted first;
      - byte order from high to low;
      - SCK frequency must at least 8 lower than system frequency.

    There are 2 types of SPI transactions:
      - incremental burst -- address increments internally after every data (array)
      - fixed burst -- one adreess, multiple data (FIFO)
    Transaction is done with 8-bit address and 16 bit data words.
    Transaction format:
      - control word (3 bytes):
          * bit 23 -- write (1) or read (0)
          * bit 22 -- burst incremental (1) or fixed (0)
          * bit 21 .. 8 -- 14 bit length (0 - 1 data word, 1 - 2 data words, etc)
          * bits 7 .. 0 -- 8 bit address
      - data word (2 bytes) 1 .. N:
          * bits 15 .. 0 -- data to be written or readen

    GPIO parameters:
      - ADBUS7 - output - active low reset for FPGA configuration (ICE_RESET)
      - BDBUS7 - output - active high reset for FPGA logic (ICE_RESET_FT)
    """

    GPIO_RESET_LOGIC_POS = 7
    GPIO_RESET_CONFIG_POS = 7

    def __init__(self, ftdi_url, spi_freq=1E6):
        """Configure the FTDI interface.

        Keyword arguments:
         ftdi_url -- device url, which can be obtained by Ftdi.show_devices()
         freq -- SPI frequency up to 8E6 (for FPGA running on 64 MHz)
        """
        # Configure SPI master
        self._spi_ctrl = SpiController()
        self._spi_ctrl.configure(ftdi_url + '2') # second port - channel B
        self._spi_port = self._spi_ctrl.get_port(cs=0, freq=spi_freq, mode=0)
        # Configure FPGA logic reset (ICE_RESET_FT)
        self._spi_gpio = self._spi_ctrl.get_gpio()
        self._spi_gpio.set_direction(1 << self.GPIO_RESET_LOGIC_POS, 1 << self.GPIO_RESET_LOGIC_POS)
        self._spi_gpio.write(0)
        # Configure FPGA configuration reset (ICE_RESET)
        self._gpio_ctrl = GpioAsyncController()
        self._gpio_ctrl.configure(ftdi_url + '1', # first port - channel A
                                  direction=(1 << self.GPIO_RESET_CONFIG_POS),
                                  frequency=1e6,
                                  initial=(1 << self.GPIO_RESET_CONFIG_POS))
        self._gpio_ctrl.write(1 << self.GPIO_RESET_CONFIG_POS)

    def _int_to_bytes(self, i, length=2):
        """Convert integer to bytes"""
        return int.to_bytes(i, length=length, byteorder='big', signed=False)

    def _words_to_bytes(self, words_list):
        """Convert list with 16 bit words to bytes"""
        bytes_str_list = [self._int_to_bytes(w, length=2) for w in words_list]
        return b''.join(bytes_str_list)  # concatenate all strings

    def _bytes_to_words(self, bytes_str):
        """Convert bytes string to list with 16 bit words"""
        return [int.from_bytes(bytes_str[b*2:b*2+2], byteorder='big', signed=False) for b in range(len(bytes_str)//2)]

    def _prepare_ctrl_word(self, addr, len, burst, wr):
        """Prepare control word for exchange.

           Keyword arguments:
             addr -- 8 bit address
             len -- number of 16 bit data words to write/read (2^14 max)
             burst -- 'fixed' address the same for every data, 'incr' - address + 1 for every next data word
             wr -- 1 for write operation, 0 - for read
        """
        ctrl_word = 0
        ctrl_word |= (wr << 23)
        ctrl_word |= ((burst == 'incr') << 22)
        ctrl_word |= (((len - 1) & 0x3FFF) << 8)
        ctrl_word |= ((addr & 0xFF) << 0)
        return ctrl_word

    def spi_read(self, addr, len=1, burst='fixed'):
        """Read data from address via SPI.

           Keyword arguments:
             addr -- 8 bit address
             len -- number of 16 bit data words to write/read (2^14 max)
             burst -- 'fixed' address the same for every data, 'incr' - address + 1 for every next data word
           Return:
             list of size 'len' with 16 bit data words
        """
        ctrl_word = self._prepare_ctrl_word(addr, len, burst, wr=0)
        rbytes = self._spi_port.exchange(self._int_to_bytes(ctrl_word, 3), len * 2)
        return self._bytes_to_words(rbytes)

    def spi_write(self, addr, data, burst='fixed'):
        """Write data to address via SPI.

           Keyword arguments:
             addr -- 8 bit address
             data -- list with 16 bit data words to write (list length 2^14 max)
             burst -- 'fixed' address the same for every data, 'incr' - address + 1 for every next data word
         """
        ctrl_word = self._prepare_ctrl_word(addr, len(data), burst, wr=1)
        wbytes = self._int_to_bytes(ctrl_word, 3) + self._words_to_bytes(data)
        self._spi_port.exchange(wbytes)

    def reset_logic_on(self):
        """Activate reset pin ICE_RESET_FT"""
        self._spi_gpio.write((1 << self.GPIO_RESET_LOGIC_POS) | self._spi_gpio.read())

    def reset_logic_off(self):
        """Deactivate reset pin ICE_RESET_FT"""
        self._spi_gpio.write(~(1 << self.GPIO_RESET_LOGIC_POS) & self._spi_gpio.read())

    def reset_config_on(self):
        """Activate reset pin ICE_RESET"""
        self._gpio_ctrl.write(~(1 << self.GPIO_RESET_CONFIG_POS) & self._gpio_ctrl.read())

    def reset_config_off(self):
        """Deactivate reset pin ICE_RESET"""
        self._gpio_ctrl.write((1 << self.GPIO_RESET_LOGIC_POS) | self._gpio_ctrl.read())

    def close_connection(self):
        """Close FTDI interface"""
        self._spi_ctrl.terminate()
        self._gpio_ctrl.close()

if __name__ == "__main__":
    Ftdi.show_devices()



'''
Wrapper for viz
'''

import numpy as np
import matplotlib.pyplot as plt
import numpy as np
import datetime

from .version import __version__


class FView(object):

    SAMPLES_PER_LINE = 16384
    GAINS_MAX = 32

    def gain_expand(self,gain,DR=False):
        if DR:
            nbPtsLine = self.SAMPLES_PER_LINE*2
        else:
            nbPtsLine = self.SAMPLES_PER_LINE
        return [gain[x // (nbPtsLine // self.GAINS_MAX)] / 1000.0 for x in range(nbPtsLine)]

    def ptstoline(self,start,stop,DR=False):
        if DR:
            nbPtsLine = self.SAMPLES_PER_LINE*2 
        else:
            nbPtsLine = self.SAMPLES_PER_LINE

        return [1 if (x >= start and x < stop) else 0 for x in range(nbPtsLine)]

    def ntons(self,N,DR=False):
        """Converts N units in nanosecs. Not impacted by DR."""
        return str(int(N/0.128))


    def readfile(self,npzPath):
        """Reads NPZ"""

        data = np.load(npzPath)
        if (data["nblines"]==32) & (data["doublerate"]==1):
            self.plotNDT(data)
        else:
            self.plotFirst(data)

        return data


    def plotFirst(self,data):
 
        t = data["t"]
        t1 = data["t_delay"]
        t2 = t1 + data["t_on"]
        t3 = t2 + data["t_inter"]
        t4 = t3 + data["t_off"]

        PON = self.ptstoline(t1,t2)
        POFF = self.ptstoline(t3,t4)
        m = int(15000//64)

        f = [k*63.75/self.SAMPLES_PER_LINE for k in range(len(data["signal"][0]))]

        FFT = np.abs(np.fft.fft(data["signal"][0]))

        plt.figure(figsize=(20,10))
        plt.subplot(2, 1, 1)

        plt.plot(t,self.gain_expand(data["gain"]),"y",label="Gain")
        plt.plot(t,data["signal"][0],"b",label="Signal")
        plt.plot(t[0:64*5],PON[0:64*5],"g",label="HV Pulse")
        plt.xlabel("us")
        plt.ylabel("Amplitude")
        plt.legend()
        title = str(data["timestamp"])+ ": "+str(data['nblines'])+ " lines.\n"
        title += "Waveform. PulseOn: "+self.ntons(data["t_on"])+"ns, damping of "+self.ntons(data["t_off"])+"ns.\n"
        title += "Python: version "+str(__version__)+". BIN: author:"+str(data["author"])+", version:"+str(data["version"])

        plt.subplot(2, 2, 3)
        plt.plot(t[0:m],data["signal"][0][0:m],alpha=0.3,label="Signal")
        plt.plot(t[0:m],PON[0:m],label="Pulse on")
        plt.plot(t[0:m],POFF[0:m], label="Dampening")
        plt.title('Pulse waveform')
        plt.ylabel('V')
        plt.ylabel('us')
        plt.legend()

        plt.subplot(2, 2, 4)
        plt.title('Spectrum composition')
        plt.plot(f[25:len(FFT)//2],FFT[25:len(FFT)//2])
        plt.xlabel('Freq (MHz)')
        plt.ylabel('Energy')

        plt.suptitle(title)
        plt.tight_layout()
        plt.savefig(str(data["nameFile"])+".jpg")
        plt.show()
        

    def plotNDT(self,data,fCentral=None,bandwidth=None):
        if (data["nblines"]==32) & (data["doublerate"]==1):

            t1 = data["t_delay"]
            t2 = t1 + data["t_on"]
            t3 = t2 + data["t_inter"]
            t4 = t3 + data["t_off"]

            PON = self.ptstoline(t1,t2,DR=True)
            POFF = self.ptstoline(t3,t4,DR=True)
            m = int(15000//64)
            Npts = self.SAMPLES_PER_LINE * 2

            even = data["signal"][0]  
            odd = data["signal"][1]  
            for k in range(15):
                even = even + data["signal"][2*(k+1)] 
                odd = odd + data["signal"][2*(k+1)+1]
                    
            signal = []
            for k in range(len(odd)):
                signal.append(odd[k])
                signal.append(even[k])
                

            t = [T*63.75*4/len(signal) for T in range(len(signal))]
            f = [2*k*63.75/Npts for k in range(Npts)]

            signal = [float(x/16.0) for x in signal]
            rawSignal = np.array(signal, copy=True)  
            FFT = np.abs(np.fft.fft(signal))
            if fCentral and bandwidth:
                FFT_clean = np.fft.fft(signal)
                N = len(f)//2
                for i in range(N+1):
                    if f[i] < (fCentral*(1-bandwidth)) or f[i]> (fCentral*(1+bandwidth)):
                        FFT_clean[i] = 0
                        FFT_clean[-i] = 0
                signal = np.real(np.fft.ifft(FFT_clean))
                msignal = np.max(np.abs(signal))
                signal = [float(x/msignal) for x in signal]

            plt.figure(figsize=(20,10))
            plt.subplot(2, 1, 1)

            plt.plot(t,self.gain_expand(data["gain"],DR=True),"y",label="Gain")
            
            if fCentral and bandwidth:
                plt.plot(t,rawSignal,"b",label="Raw signal" ,alpha = 0.3)
                plt.plot(t,signal,"r",label="Filtered signal (normalized)")
            else:
                plt.plot(t,signal,"b",label="Signal")

            plt.plot(t[0:64*5],PON[0:64*5],"g",label="HV Pulse")
            plt.xlim(0, 50) 
            plt.xlabel("us")
            plt.ylabel("Amplitude")
            plt.legend()
            title = str(data["timestamp"])+ ": "+str(data['nblines'])+ " lines.\n"
            title += "NDT averaged waveform (first 50us). PulseOn: "+self.ntons(data["t_on"])+"ns, damping of "+self.ntons(data["t_off"])+"ns.\n"
            title += "Python: saved version "+str(__version__)+" (saved with "+str(data["libversion"])+"). BIN: author:"+str(data["author"])+", version:"+str(data["version"])

            plt.subplot(2, 2, 3)
            if fCentral and bandwidth:
                plt.plot(t[0:m],rawSignal[0:m],"b",label="Raw signal" ,alpha = 0.1)
                plt.plot(t[0:m],signal[0:m],"r",label="Filtered signal (normalized)",alpha=0.3)
            else:
                plt.plot(t[0:m],signal[0:m],"b",label="Signal") 
            plt.plot(t[0:m],PON[0:m],label="Pulse on")
            plt.plot(t[0:m],POFF[0:m], label="Dampening")
            plt.title('Pulse waveform')
            plt.ylabel('V')
            plt.ylabel('us')
            plt.legend()

            plt.subplot(2, 2, 4)
            plt.title('Spectrum composition')
            plt.plot(f[50:len(FFT)//2],FFT[50:len(FFT)//2])
            if fCentral:
                if bandwidth:
                    plt.plot(f[50:len(FFT_clean)//2],np.abs(FFT_clean)[50:len(FFT_clean)//2])
            plt.xlabel('Freq (MHz)')
            plt.ylabel('Energy')

            plt.suptitle(title)
            plt.tight_layout()
            plt.savefig(str(data["nameFile"])+"_ndt.jpg")
            plt.show()

        else:
            print("Conditions - 32 lines & double rate - not met for this plot.")
     





class CsrMap:
    """Control/Status register map"""

    # INITDEL - Initial pulse delay -- 0 - 1 period of 127.5 MHz, 1 - 2 periods, etc.
    INITDEL_ADDR = 0x00
    INITDEL_WIDTH = 8
    INITDEL_MASK = 0xff

    # PONW - Pon width -- 0 - 1 period of 127.5 MHz, 1 - 2 periods, etc.
    PONW_ADDR = 0x01
    PONW_WIDTH = 8
    PONW_MASK = 0xff

    # POFFW - Poff width -- 0 - 1 period of 127.5 MHz, 1 - 2 periods, etc.
    POFFW_ADDR = 0x02
    POFFW_WIDTH = 8
    POFFW_MASK = 0xff

    # INTERW - Intermediate delay width -- 0 - 1 period of 127.5 MHz, 1 - 2 periods, etc.
    INTERW_ADDR = 0x03
    INTERW_WIDTH = 8
    INTERW_MASK = 0xff

    # DRMODE - Double resolution mode -- add 1 to INITDEL whe line is even
    DRMODE_ADDR = 0x04
    DRMODE_WIDTH = 1
    DRMODE_MASK = 0x1

    # DACOUT - DAC out -- value for the DAC idle state
    DACOUT_ADDR = 0x07
    DACOUT_WIDTH = 10
    DACOUT_MASK = 0x3ff

    # DACGAIN - DAC gain
    DACGAIN_ADDR = 0x20
    DACGAIN_WIDTH = 10
    DACGAIN_MASK = 0x3ff
    DACGAIN_N = 32

    # ACQSTART - Start acquisition
    ACQSTART_ADDR = 0x50
    ACQSTART_WIDTH = 1
    ACQSTART_MASK = 0x1

    # ACQDONE - Acquisition is done
    ACQDONE_ADDR = 0x51
    ACQDONE_WIDTH = 1
    ACQDONE_MASK = 0x1

    # NBLINES - Number of lines per acquisition -- 0 - 1 line, 1 - 2 lines, etc.
    NBLINES_ADDR = 0x52
    NBLINES_WIDTH = 8
    NBLINES_MASK = 0xff

    # ACQBUSY - Acquisition is busy
    ACQBUSY_ADDR = 0x53
    ACQBUSY_WIDTH = 1
    ACQBUSY_MASK = 0x1

    # LED1 - LED1 (LED_ACQUISITION) control
    LED1_ADDR = 0x63
    LED1_WIDTH = 1
    LED1_MASK = 0x1

    # LED2 - LED2 (LED_SiNGLE/nLOOP) control
    LED2_ADDR = 0x64
    LED2_WIDTH = 1
    LED2_MASK = 0x1

    # LED3 - LED3 control
    LED3_ADDR = 0x65
    LED3_WIDTH = 1
    LED3_MASK = 0x1

    # TOPTURN1 - TOP_TURN1 status
    TOPTURN1_ADDR = 0x66
    TOPTURN1_WIDTH = 1
    TOPTURN1_MASK = 0x1

    # TOPTURN2 - TOP_TURN2 status
    TOPTURN2_ADDR = 0x67
    TOPTURN2_WIDTH = 1
    TOPTURN2_MASK = 0x1

    # TOPTURN3 - TOP_TURN3 status
    TOPTURN3_ADDR = 0x68
    TOPTURN3_WIDTH = 1
    TOPTURN3_MASK = 0x1

    # JUMPER1 - Jumper1 status
    JUMPER1_ADDR = 0x69
    JUMPER1_WIDTH = 1
    JUMPER1_MASK = 0x1

    # JUMPER2 - Jumper2 status
    JUMPER2_ADDR = 0x6A
    JUMPER2_WIDTH = 1
    JUMPER2_MASK = 0x1

    # JUMPER3 - Jumper3 status
    JUMPER3_ADDR = 0x6B
    JUMPER3_WIDTH = 1
    JUMPER3_MASK = 0x1

    # OUT1ICE - OUT1_ICE output control
    OUT1ICE_ADDR = 0x6C
    OUT1ICE_WIDTH = 1
    OUT1ICE_MASK = 0x1

    # OUT2ICE - OUT2_ICE output control
    OUT2ICE_ADDR = 0x6D
    OUT2ICE_WIDTH = 1
    OUT2ICE_MASK = 0x1

    # OUT3ICE - OUT3_ICE output control
    OUT3ICE_ADDR = 0x6E
    OUT3ICE_WIDTH = 1
    OUT3ICE_MASK = 0x1

    # RAMDATA - Read data from the external RAM
    RAMDATA_ADDR = 0xA0
    RAMDATA_WIDTH = 16
    RAMDATA_MASK = 0xffff
    RAMDATA_N = 16384

    # RAMRADDRRST - Reset external RAM read address
    RAMRADDRRST_ADDR = 0xA1
    RAMRADDRRST_WIDTH = 1
    RAMRADDRRST_MASK = 0x1

    # RAMFINC - Fill external RAM with incrementing data pattern
    RAMFINC_ADDR = 0xA4
    RAMFINC_WIDTH = 1
    RAMFINC_MASK = 0x1

    # RAMFDEC - Fill external RAM with decrementing data pattern
    RAMFDEC_ADDR = 0xA5
    RAMFDEC_WIDTH = 1
    RAMFDEC_MASK = 0x1

    # RAMFDONE - Filling of external RAM is done
    RAMFDONE_ADDR = 0xA6
    RAMFDONE_WIDTH = 1
    RAMFDONE_MASK = 0x1

    # AUTHOR - Author
    AUTHOR_ADDR = 0xF0
    AUTHOR_WIDTH = 8
    AUTHOR_MASK = 0xff

    # VERSION - Version
    VERSION_ADDR = 0xF1
    VERSION_WIDTH = 8
    VERSION_MASK = 0xff

    def __init__(self, ftdidev):
        self._ftdi = ftdidev

    @property
    def initdel(self):
        """Get current INITDEL register value"""
        data = self._ftdi.spi_read(self.INITDEL_ADDR, len=1, burst='fixed')
        return data[0] & self.INITDEL_MASK

    @initdel.setter
    def initdel(self, val):
        """Set INITDEL register with new value"""
        data = val & self.INITDEL_MASK
        self._ftdi.spi_write(self.INITDEL_ADDR, [data], burst='fixed')

    @property
    def ponw(self):
        """Get current PONW register value"""
        data = self._ftdi.spi_read(self.PONW_ADDR, len=1, burst='fixed')
        return data[0] & self.PONW_MASK

    @ponw.setter
    def ponw(self, val):
        """Set PONW register with new value"""
        data = val & self.PONW_MASK
        self._ftdi.spi_write(self.PONW_ADDR, [data], burst='fixed')

    @property
    def poffw(self):
        """Get current POFFW register value"""
        data = self._ftdi.spi_read(self.POFFW_ADDR, len=1, burst='fixed')
        return data[0] & self.POFFW_MASK

    @poffw.setter
    def poffw(self, val):
        """Set POFFW register with new value"""
        data = val & self.POFFW_MASK
        self._ftdi.spi_write(self.POFFW_ADDR, [data], burst='fixed')

    @property
    def interw(self):
        """Get current INTERW register value"""
        data = self._ftdi.spi_read(self.INTERW_ADDR, len=1, burst='fixed')
        return data[0] & self.INTERW_MASK

    @interw.setter
    def interw(self, val):
        """Set INTERW register with new value"""
        data = val & self.INTERW_MASK
        self._ftdi.spi_write(self.INTERW_ADDR, [data], burst='fixed')

    @property
    def drmode(self):
        """Get current DRMODE register value"""
        data = self._ftdi.spi_read(self.DRMODE_ADDR, len=1, burst='fixed')
        return data[0] & self.DRMODE_MASK

    @drmode.setter
    def drmode(self, val):
        """Set DRMODE register with new value"""
        data = val & self.DRMODE_MASK
        self._ftdi.spi_write(self.DRMODE_ADDR, [data], burst='fixed')

    @property
    def dacout(self):
        """Get current DACOUT register value"""
        data = self._ftdi.spi_read(self.DACOUT_ADDR, len=1, burst='fixed')
        return data[0] & self.DACOUT_MASK

    @dacout.setter
    def dacout(self, val):
        """Set DACOUT register with new value"""
        data = val & self.DACOUT_MASK
        self._ftdi.spi_write(self.DACOUT_ADDR, [data], burst='fixed')

    @property
    def dacgain(self):
        """Get current DACGAIN registers values"""
        data = self._ftdi.spi_read(self.DACGAIN_ADDR, len=self.DACGAIN_N, burst='incr')
        return [w & self.DACGAIN_MASK for w in data]

    @dacgain.setter
    def dacgain(self, val):
        """Set DACGAIN registers with new values"""
        data = [w & self.DACGAIN_MASK for w in val]
        self._ftdi.spi_write(self.DACGAIN_ADDR, data, burst='incr')

    @property
    def acqstart(self):
        """Get current ACQSTART register value"""
        return 0

    @acqstart.setter
    def acqstart(self, val):
        """Set ACQSTART register with new value"""
        data = val & self.ACQSTART_MASK
        self._ftdi.spi_write(self.ACQSTART_ADDR, [data], burst='fixed')

    @property
    def acqdone(self):
        """Get current ACQDONE register value"""
        data = self._ftdi.spi_read(self.ACQDONE_ADDR, len=1, burst='fixed')
        return data[0] & self.ACQDONE_MASK

    @property
    def nblines(self):
        """Get current NBLINES register value"""
        data = self._ftdi.spi_read(self.NBLINES_ADDR, len=1, burst='fixed')
        return data[0] & self.NBLINES_MASK

    @nblines.setter
    def nblines(self, val):
        """Set NBLINES register with new value"""
        data = val & self.NBLINES_MASK
        self._ftdi.spi_write(self.NBLINES_ADDR, [data], burst='fixed')

    @property
    def acqbusy(self):
        """Get current ACQBUSY register value"""
        data = self._ftdi.spi_read(self.ACQBUSY_ADDR, len=1, burst='fixed')
        return data[0] & self.ACQBUSY_MASK

    @property
    def led1(self):
        """Get current LED1 register value"""
        data = self._ftdi.spi_read(self.LED1_ADDR, len=1, burst='fixed')
        return data[0] & self.LED1_MASK

    @led1.setter
    def led1(self, val):
        """Set LED1 register with new value"""
        data = val & self.LED1_MASK
        self._ftdi.spi_write(self.LED1_ADDR, [data], burst='fixed')

    @property
    def led2(self):
        """Get current LED2 register value"""
        data = self._ftdi.spi_read(self.LED2_ADDR, len=1, burst='fixed')
        return data[0] & self.LED2_MASK

    @led2.setter
    def led2(self, val):
        """Set LED2 register with new value"""
        data = val & self.LED2_MASK
        self._ftdi.spi_write(self.LED2_ADDR, [data], burst='fixed')

    @property
    def led3(self):
        """Get current LED3 register value"""
        data = self._ftdi.spi_read(self.LED3_ADDR, len=1, burst='fixed')
        return data[0] & self.LED3_MASK

    @led3.setter
    def led3(self, val):
        """Set LED3 register with new value"""
        data = val & self.LED3_MASK
        self._ftdi.spi_write(self.LED3_ADDR, [data], burst='fixed')

    @property
    def topturn1(self):
        """Get current TOPTURN1 register value"""
        data = self._ftdi.spi_read(self.TOPTURN1_ADDR, len=1, burst='fixed')
        return data[0] & self.TOPTURN1_MASK

    @property
    def topturn2(self):
        """Get current TOPTURN2 register value"""
        data = self._ftdi.spi_read(self.TOPTURN2_ADDR, len=1, burst='fixed')
        return data[0] & self.TOPTURN2_MASK

    @property
    def topturn3(self):
        """Get current TOPTURN3 register value"""
        data = self._ftdi.spi_read(self.TOPTURN3_ADDR, len=1, burst='fixed')
        return data[0] & self.TOPTURN3_MASK

    @property
    def jumper1(self):
        """Get current JUMPER1 register value"""
        data = self._ftdi.spi_read(self.JUMPER1_ADDR, len=1, burst='fixed')
        return data[0] & self.JUMPER1_MASK

    @property
    def jumper2(self):
        """Get current JUMPER2 register value"""
        data = self._ftdi.spi_read(self.JUMPER2_ADDR, len=1, burst='fixed')
        return data[0] & self.JUMPER2_MASK

    @property
    def jumper3(self):
        """Get current JUMPER3 register value"""
        data = self._ftdi.spi_read(self.JUMPER3_ADDR, len=1, burst='fixed')
        return data[0] & self.JUMPER3_MASK

    @property
    def out1ice(self):
        """Get current OUT1ICE register value"""
        data = self._ftdi.spi_read(self.OUT1ICE_ADDR, len=1, burst='fixed')
        return data[0] & self.OUT1ICE_MASK

    @out1ice.setter
    def out1ice(self, val):
        """Set OUT1ICE register with new value"""
        data = val & self.OUT1ICE_MASK
        self._ftdi.spi_write(self.OUT1ICE_ADDR, [data], burst='fixed')

    @property
    def out2ice(self):
        """Get current OUT2ICE register value"""
        data = self._ftdi.spi_read(self.OUT2ICE_ADDR, len=1, burst='fixed')
        return data[0] & self.OUT2ICE_MASK

    @out2ice.setter
    def out2ice(self, val):
        """Set OUT2ICE register with new value"""
        data = val & self.OUT2ICE_MASK
        self._ftdi.spi_write(self.OUT2ICE_ADDR, [data], burst='fixed')

    @property
    def out3ice(self):
        """Get current OUT3ICE register value"""
        data = self._ftdi.spi_read(self.OUT3ICE_ADDR, len=1, burst='fixed')
        return data[0] & self.OUT3ICE_MASK

    @out3ice.setter
    def out3ice(self, val):
        """Set OUT3ICE register with new value"""
        data = val & self.OUT3ICE_MASK
        self._ftdi.spi_write(self.OUT3ICE_ADDR, [data], burst='fixed')

    @property
    def ramdata(self):
        """Get all data from RAMDATA buffer"""
        data = self._ftdi.spi_read(self.RAMDATA_ADDR, len=self.RAMDATA_N, burst='fixed')
        return [w & self.RAMDATA_MASK for w in data]

    @property
    def ramraddrrst(self):
        """Get current RAMRADDRRST register value"""
        return 0

    @ramraddrrst.setter
    def ramraddrrst(self, val):
        """Set RAMRADDRRST register with new value"""
        data = val & self.RAMRADDRRST_MASK
        self._ftdi.spi_write(self.RAMRADDRRST_ADDR, [data], burst='fixed')

    @property
    def ramfinc(self):
        """Get current RAMFINC register value"""
        return 0

    @ramfinc.setter
    def ramfinc(self, val):
        """Set RAMFINC register with new value"""
        data = val & self.RAMFINC_MASK
        self._ftdi.spi_write(self.RAMFINC_ADDR, [data], burst='fixed')

    @property
    def ramfdec(self):
        """Get current RAMFDEC register value"""
        return 0

    @ramfdec.setter
    def ramfdec(self, val):
        """Set RAMFDEC register with new value"""
        data = val & self.RAMFDEC_MASK
        self._ftdi.spi_write(self.RAMFDEC_ADDR, [data], burst='fixed')

    @property
    def ramfdone(self):
        """Get current RAMFDONE register value"""
        data = self._ftdi.spi_read(self.RAMFDONE_ADDR, len=1, burst='fixed')
        return data[0] & self.RAMFDONE_MASK

    @property
    def author(self):
        """Get current AUTHOR register value"""
        data = self._ftdi.spi_read(self.AUTHOR_ADDR, len=1, burst='fixed')
        return data[0] & self.AUTHOR_MASK

    @property
    def version(self):
        """Get current VERSION register value"""
        data = self._ftdi.spi_read(self.VERSION_ADDR, len=1, burst='fixed')
        return data[0] & self.VERSION_MASK



from un0usb.csr_map import CsrMap
from un0usb.fpga_ctrl import FpgaControl, Acquisition
from un0usb.ftdi_dev import FtdiDevice
from un0usb.viz import FView
from .version import __version__

__author__ = "kelu124"


#from distutils.core import setup
from setuptools import setup
#import glob, os, re, sys
#import setuptools; setuptools.bootstrap_install_from = egg
#from setuptools import Extension, find_packages, setup
# read the contents of your README file


exec(open('un0usb/version.py').read())

setup(
  name = 'un0usb',         # How you named your package folder (MyLib)
  packages = ['un0usb'],   # Chose the same as "name"
  version = __version__,      # Start with a small number and increase it with every change you make
  license='gpl-3.0',        # Chose a license from here: https://help.github.com/articles/licensing-a-repository
  description = 'Library for un0rick usb interface',   # Give a short description about your library
  author = 'Luc Jonveaux',                   # Type in your name
  author_email = 'kelu124@gmail.com',      # Type in your E-Mail
  long_description=open('README.md').read(),
  long_description_content_type='text/markdown',
  url = 'http://un0rick.cc',   # Provide either the link to your github or to your website
  download_url = 'https://github.com/kelu124/python-usb-un0rick/archive/v0.2.1.tar.gz',    # I explain this later on
  keywords = ['ultrasound', 'usb', 'un0rick'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'pyftdi','numpy','matplotlib',
          ],
  classifiers=[
    'Development Status :: 3 - Alpha',      # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
    'Intended Audience :: Science/Research',      # Define that your audience are developers
    'Topic :: Scientific/Engineering :: Physics',
    'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',   # Again, pick a license
    'Programming Language :: Python :: 3',      #Specify which pyhton versions that you want to support
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
  ],
  zip_safe=False,
  include_package_data=True, 
  data_files=[('firmware',['usb.bin',"README.md"]),],
)



'''
Main FPGA control script
'''

import numpy as np
import datetime

from time import sleep
from un0usb.csr_map import CsrMap
from un0usb.ftdi_dev import FtdiDevice
from .version import __version__

class FpgaControl(object):
    """Collection of FPGA control functions via FTDI API"""
    MAX_LINES = 32
    WORDS_PER_LINE = 16384

    def __init__(self, ftdi_url, spi_freq=1E6):
        """Initialize FPGA controller.

        Keyword arguments:
         ftdi_url -- FTDI device url, which can be obtained by Ftdi.show_devices()
         freq -- SPI frequency up to 8E6 (for FPGA running on 64 MHz)
        """
        self._ftdi = FtdiDevice(ftdi_url, spi_freq)
        self.csr = CsrMap(self._ftdi)

    def reset(self):
        """Reset FPGA logic"""
        self._ftdi.reset_logic_on()
        sleep(0.01)
        self._ftdi.reset_logic_off()

    def reload(self):
        """Reload FPGA configuration from flash"""
        self._ftdi.reset_config_on()
        sleep(0.01)
        self._ftdi.reset_config_off()
        sleep(0.5)

    def read_lines(self, n):
        """Read 'n' number of lines from SRAM buffer

        Maxinum 'n' -- 32
        """
        # reset external ram address to read from the memory beginning
        self.csr.ramraddrrst = 1
        res = []
        for _ in range(n):
            # read lines (16384 words per line) one by one
            line = self.csr.ramdata
            res += [line]
        return res

    def set_pulseform(self, initDelay=5, POn=16, PInter=16, Poff=5000):
        """Set pulser.
        8 ~ 42ns
        16 ~ 167ns
        25 ~ 208ns
        32 ~ 250ns
        50 units ~ 420ns
        100 ~ 794ns
        250 ~  1.5us
        1000 ~ 1833ns
        2500 ~ 1.5us
        Keyword arguments:
          initDelay -- ncycles before acquisition starts
          POn -- width of the pulse
          PInter -- time between Pon and PDamp
          Poff -- damping period
        """
        if initDelay:
            self.csr.initdel= initDelay
        if POn:
            self.csr.ponw = POn
        if PInter:
            self.csr.interw = PInter
        if Poff:
            self.csr.poffw = Poff

    def stdNDTacq(self):
        """Do standard acquisition - 32lines, interleaved, standard gain.
        """
        self.do_acquisition(acq_lines=32, gain=None, double_rate=True)
        now = datetime.datetime.today().strftime('%Y%m%d%H%M%S')
        return self.save(nameFile = now+"_ndt")

    def do_acquisition(self, acq_lines=1, gain=None, double_rate=False):
        """Do acquisitions.
        
        Keyword arguments:
          acq_lines -- number of lines to sample: int 1 .. 32
          gain -- list with gain values: None or list with length of 32
          double_rate -- enable/disable interleaving mode: bool
        """
        if gain:
            self.csr.dacgain = gain
        else:
            gain = [int(100 + ((1000-100)*x*x*x/32/32/32)) for x in range(32)]
            self.csr.dacgain = gain
        self.csr.nblines = acq_lines - 1
        self.csr.drmode = int(double_rate)
        self.csr.acqstart = 1
        while (not self.csr.acqdone):
            sleep(0.01)
        return self.read_lines(acq_lines)

    def disconnect(self):
        """Disconnect from FTDI and close all open ports"""
        self._ftdi.close_connection()

    def line_to_voltage(self, line):
        """Extracting voltage reading from line raw data"""
        SAMPLE_W = 10
        SAMPLE_N = 2 ** SAMPLE_W
        res = [((2 * 1.0) / SAMPLE_N) * ((w & (SAMPLE_N - 1)) - SAMPLE_N // 2) for w in line]
        return np.array(res)

    def save(self, nameFile = None):
        """Saving acquisition in npz"""
        acq_res = self.read_lines( self.csr.nblines + 1 )
        allAcqs = []
        for k in range(len(acq_res)):
            allAcqs.append( self.line_to_voltage(acq_res[k]) )
        
        t = [x*256.0/len(acq_res[0]) for x in range(len(acq_res[0]))]

        now = datetime.datetime.today().strftime('%Y%m%d%H%M%S')
        if nameFile:
            nameFile = now

        np.savez_compressed(nameFile, signal=allAcqs, t=t, nblines=int(self.csr.nblines+1), 
                            gain=self.csr.dacgain, t_on = self.csr.ponw, 
                            dac = self.csr.dacout,
                            t_inter = self.csr.interw,
                            t_off = self.csr.poffw,
                            t_delay = self.csr.initdel,
                            author = self.csr.author,
                            version = self.csr.version,
                            doublerate = self.csr.drmode,
                            libversion = str(__version__),
                            timestamp = str(now),
                            nameFile = str(nameFile) )

        return nameFile+".npz"


class Acquisition(object):
    def empty():
        return False

if __name__ == "__main__":
    # init FTDI device
    fpga = FpgaControl('ftdi://ftdi:2232:/', spi_freq=8E6)

    # reload configuration (optional step - just to fill BRAM (DACGAIN registers) with initial values)
    fpga.reload()

    # reset fpga
    fpga.reset()

    # read initial state of INITDEL register
    print("initdel = 0x%x" % fpga.csr.initdel)
    # write new value to the INITDEL
    fpga.csr.initdel = 0x20
    # read current state of INITDEL register
    print("initdel = 0x%x" % fpga.csr.initdel)

    # read DACGAIN array initial state
    print("dacgain = ", fpga.csr.dacgain)
    # write new values to the DACGAIN
    fpga.csr.dacgain = [200 + i for i in range(32)]
    # read DACGAIN array current state
    print("dacgain = ", fpga.csr.dacgain)

    # some LED3 blinking
    fpga.csr.led3 = 1
    sleep(1)
    fpga.csr.led3 = 0
    sleep(0.3)
    fpga.csr.led3 = 1
    sleep(0.3)
    fpga.csr.led3 = 0

    # reset fpga again (optional)
    fpga.reset()

    # close FTDI interface
    fpga.disconnect()
